#define _USE_MATH_DEFINES true
#include <math.h>
#include "Camera.h"
#define PASSO 0.5


//***************************************************
// Construtores
//
Camera::Camera()
{
    camera.x = 0.0;
    camera.y = 0.0;
    camera.z = -5;
    
    foco.x = 0.0;
    foco.y = 0.0;
    foco.z = 0.0;
    
    view.x = 0.0;
    view.y = 1.0;
    view.z = 0.0;    
    AtualizarPosicao();
}

Camera::Camera(Ponto3D camera, Ponto3D foco, Ponto3D view)
{
    Camera::camera = camera;
    Camera::foco = foco;
    Camera::view = view;
    AtualizarPosicao();
}
//
// Fim Construtores
//***************************************************

//***************************************************
// Operaes Cmera
//
void Camera::Azimuth(double ang)
{
    double dx,dy,dz;
    double t;
    double camx, camy, camz;
    t = (ang * M_PI * 2) / 360;
    dx = - foco.x;
    dy = - foco.y;
    dz = - foco.z;
    
    camera.x += dx;
    camera.y += dy;
    camera.z += dz;    
    
    camx = camera.x;
    camy = camera.y;
    camz = camera.z;
   
    // faz a rotacao em x
    if (nzero(view.x))
    {
       // camera.x = view.x * camera.x; //0
        camera.y = cos(t)*camy + (-sin(t))*camz; // 0
        camera.z = sin(t)*camy + cos(t)*camz; // -2.5
        camx = camera.x;
        camy = camera.y;
        camz = camera.z;
    }    
    
    // faz a rotacao em y
    if (nzero(view.y))
    {
        camera.x = cos(t)*camx + sin(t)*camz; // 0
      //  camera.y = view.y * camera.y; // 0
        camera.z = cos(t)*camz -sin(t)*camx; // -2.5    
        camx = camera.x;
        camy = camera.y;
        camz = camera.z;
    }
    
    // faz a rotacao em z
    if (nzero(view.z))
    {
        camera.x = cos(t)*camx + (-sin(t))*camy; // 0
        camera.y = sin(t)*camx + cos(t)*camy; // 0
    //    camera.z = view.z*camera.z; // 0
    }
    
    camera.x -= dx;
    camera.y -= dy;
    camera.z -= dz;
}

void Camera::Elevation(double ang)
{
    Ponto3D ref;
    double dx,dy,dz;
    double t;
    double camx, camy, camz;
    double vx, vy, vz;
    t = (ang * M_PI * 2) / 360;
    dx = - foco.x;
    dy = - foco.y;
    dz = - foco.z;
    
    camera.x += dx;
    camera.y += dy;
    camera.z += dz; 
    
    ref.x = view.y * camera.z - view.z * camera.y;        
    ref.y = view.z * camera.x - view.x * camera.z ;
    ref.z = view.x * camera.y - view.y * camera.x;
  
/*    printf("ref x: %f y: %f z: %f\n", ref.x, ref.y, ref.z);*/
    
    camx = camera.x;
    camy = camera.y;
    camz = camera.z;
   
    // faz a rotacao em x
    if (nzero(ref.x))
    {
        camera.y = cos(t)*camy + (-sin(t))*camz;
        camera.z = sin(t)*camy + cos(t)*camz;
        camx = camera.x;
        camy = camera.y;
        camz = camera.z;
    }    
    
    // faz a rotacao em y
    if (nzero(ref.y))
    {
        camera.x = cos(t)*camx + sin(t)*camz;
        camera.z = cos(t)*camz -sin(t)*camx;
        camx = camera.x;
        camy = camera.y;
        camz = camera.z;
    }
    
    // faz a rotacao em z
    if (nzero(ref.z))
    {
        camera.x = cos(t)*camx + (-sin(t))*camy;
        camera.y = sin(t)*camx + cos(t)*camy;
    }
    
    vx = view.x;
    vy = view.y;
    vz = view.z;
    // rotacao do view
    // faz a rotacao em x
    if (nzero(ref.x))
    {
        view.y = cos(t)*vy + (-sin(t))*vz;
        view.z = sin(t)*vy + cos(t)*vz;
        vx = view.x;
        vy = view.y;
        vz = view.z;
    }    
    
    // faz a rotacao em y
    if (nzero(ref.y))
    {
        view.x = cos(t)*vx + sin(t)*vz;
        view.z = cos(t)*vz -sin(t)*vx;
        vx = view.x;
        vy = view.y;
        vz = view.z;
    }
    
    // faz a rotacao em z
    if (nzero(ref.z))
    {
        view.x = cos(t)*vx + (-sin(t))*vy;
        view.y = sin(t)*vx + cos(t)*vy;
    }
    
    camera.x -= dx;
    camera.y -= dy;
    camera.z -= dz;
}

void Camera::Pitch(double ang)
{
    Ponto3D ref;
    double dx,dy,dz;
    double t;
    double fx, fy, fz;
    double vx, vy, vz;
    t = (ang * M_PI * 2) / 360;
    dx = - camera.x;
    dy = - camera.y;
    dz = - camera.z;
    
    // traz o foco para referencia 
    foco.x += dx;
    foco.y += dy;
    foco.z += dz; 
    
    ref.x = view.y * foco.z - view.z * foco.y;        
    ref.y = view.z * foco.x - view.x * foco.z;
    ref.z = view.x * foco.y - view.y * foco.x;
    
    fx = foco.x;
    fy = foco.y;
    fz = foco.z;
   
    // faz a rotacao em x
    if (nzero(ref.x))
    {
        foco.y = cos(t)*fy + (-sin(t))*fz;
        foco.z = sin(t)*fy + cos(t)*fz;
        fx = foco.x;
        fy = foco.y;
        fz = foco.z;
    }    
    
    // faz a rotacao em y
    if (nzero(ref.y))
    {
        foco.x = cos(t)*fx + sin(t)*fz;
        foco.z = cos(t)*fz -sin(t)*fx;
        fx = foco.x;
        fy = foco.y;
        fz = foco.z;
    }
    
    // faz a rotacao em z
    if (nzero(ref.z))
    {
        foco.x = cos(t)*fx + (-sin(t))*fy;
        foco.y = sin(t)*fx + cos(t)*fy;
    }
    
    vx = view.x;
    vy = view.y;
    vz = view.z;
    // rotacao do view
    // faz a rotacao em x
    if (nzero(ref.x))
    {
        view.y = cos(t)*vy + (-sin(t))*vz;
        view.z = sin(t)*vy + cos(t)*vz;
        vx = view.x;
        vy = view.y;
        vz = view.z;
    }    
    
    // faz a rotacao em y
    if (nzero(ref.y))
    {
        view.x = cos(t)*vx + sin(t)*vz;
        view.z = cos(t)*vz -sin(t)*vx;
        vx = view.x;
        vy = view.y;
        vz = view.z;
    }
    
    // faz a rotacao em z
    if (nzero(ref.z))
    {
        view.x = cos(t)*vx + (-sin(t))*vy;
        view.y = sin(t)*vx + cos(t)*vy;
    }
    
    foco.x -= dx;
    foco.y -= dy;
    foco.z -= dz;
}

void Camera::Roll(double ang)
{
    double dx,dy,dz;
    double t;
    double vx, vy, vz;
    t = (ang * M_PI * 2) / 360;
    dx = - foco.x;
    dy = - foco.y;
    dz = - foco.z;
    
    camera.x += dx;
    camera.y += dy;
    camera.z += dz;    
    
    vx = view.x;
    vy = view.y;
    vz = view.z;
   
    // faz a rotacao em x
    if (nzero(camera.x))
    {
        view.y = cos(t)*vy + (-sin(t))*vz;
        view.z = sin(t)*vy + cos(t)*vz;
        vx = view.x;
        vy = view.y;
        vz = view.z;
    }    
    
    // faz a rotacao em y
    if (nzero(camera.y))
    {
        view.x = cos(t)*vx + sin(t)*vz;
        view.z = cos(t)*vz -sin(t)*vx;
        vx = view.x;
        vy = view.y;
        vz = view.z;
    }
    
    // faz a rotacao em z
    if (nzero(camera.z))
    {
        view.x = cos(t)*vx + (-sin(t))*vy;
        view.y = sin(t)*vx + cos(t)*vy;
    }
    
    camera.x -= dx;
    camera.y -= dy;
    camera.z -= dz;
}

void Camera::Yaw(double ang)
{
    double dx,dy,dz;
    double t;
    double fx, fy, fz;
    t = (ang * M_PI * 2) / 360;
    dx = - camera.x;
    dy = - camera.y;
    dz = - camera.z;
    
    foco.x += dx;
    foco.y += dy;
    foco.z += dz;    
    
    fx = foco.x;
    fy = foco.y;
    fz = foco.z;
   
    // faz a rotacao em x
    if (nzero(view.x))
    {
        foco.y = cos(t)*fy + (-sin(t))*fz; // 0
        foco.z = sin(t)*fy + cos(t)*fz; // -2.5
        fx = foco.x;
        fy = foco.y;
        fz = foco.z;
    }    
    
    // faz a rotacao em y
    if (nzero(view.y))
    {
        foco.x = cos(t)*fx + sin(t)*fz; // 0
        foco.z = cos(t)*fz -sin(t)*fx; // -2.5 
        fx = foco.x;
        fy = foco.y;
        fz = foco.z;
    }
    
    // faz a rotacao em z
    if (nzero(view.z))
    {
        foco.x = cos(t)*fx + (-sin(t))*fy; // 0
        foco.y = sin(t)*fx + cos(t)*fy; // 0
    }
    
    foco.x -= dx;
    foco.y -= dy;
    foco.z -= dz;
}

void Camera::Translate(double tx, double ty, double tz)
{
    camera.x += tx;
    camera.y += ty;
    camera.z += tz;
    
    foco.x += tx;
    foco.y += ty;
    foco.z += tz;        
    
    //AtualizarPosicao();
}

void Camera::AtualizarPosicao()
{
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(45.0f, 1, 0.1f, 100.0f);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity(); 
    gluLookAt(camera.x, camera.y, camera.z,
                foco.x,   foco.y,   foco.z,
                view.x,   view.y,   view.z);
    /*printf("camera x: %f y: %f z: %f\n", camera.x, camera.y, camera.z);
    printf("foco x: %f y: %f z: %f\n", foco.x, foco.y, foco.z);
    printf("view x: %f y: %f z: %f\n", view.x, view.y, view.z);
    printf("--------------------------------\n");*/
}

void Camera::moveUp(){
    camera.y+=PASSO;
    foco.y+=PASSO;
}

void Camera::moveDown(){
    camera.y-=PASSO;
    foco.y-=PASSO;
}

void Camera::moveFront(){
    camera.z+=PASSO;
    foco.z+=PASSO;
}

void Camera::moveBack(){
    camera.z-=PASSO;
    foco.z-=PASSO;
} 

void Camera::moveLeft(){
    camera.x+=PASSO;
    foco.x+=PASSO;
}

void Camera::moveRight(){
    camera.x-=PASSO;
    foco.x-=PASSO;
}

//
// Fim Operaes Cmera
//***************************************************
bool Camera::nzero(double val)
{
    double modval;
    modval = val - 0;
    if (modval < 0) modval = modval * -1;
    if (modval < exp(-3.0))
        return false;
    return true;
}
